/**
 * @vue-async/resource-manager@0.1.1-alpha.3
 */

(function (global, factory) {
    typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('vue')) :
    typeof define === 'function' && define.amd ? define(['exports', 'vue'], factory) :
    (global = global || self, factory(global['async-manager'] = {}, global.Vue));
}(this, (function (exports, Vue) { 'use strict';

    Vue = Vue && Object.prototype.hasOwnProperty.call(Vue, 'default') ? Vue['default'] : Vue;

    /**
     * @vue-async/utils@0.1.3-alpha.3
     */
    function error(condition, msg, vm) {
        if (!condition) {
            // eslint-disable-next-line no-console
            console.error("[@vue-async] " + msg, vm);
        }
    }

    var currentInstance = null;
    /**
     * set current instance
     * @param instance Vue instance
     */
    function setCurrentInstance(instance) {
        currentInstance = instance;
    }
    var suspenseInstanceStack = [];
    var currentSuspenseInstance;
    /**
     * push suspense instance
     * @param instance Vue instance
     */
    function pushSuspenseInstance(instance) {
        currentSuspenseInstance = instance;
        suspenseInstanceStack.push(instance);
    }
    /**
     * pop suspense instance
     */
    function popSuspenseInstance() {
        suspenseInstanceStack.pop();
        return (currentSuspenseInstance = suspenseInstanceStack[suspenseInstanceStack.length - 1]);
    }

    var RESOLVED = 'resolved';
    var REJECTED = 'rejected';
    var COMPONENT_NAME = '__VueSuspense__';
    var del = function (asyncFactory, error) {
        var suspIns = asyncFactory.suspenseInstance;
        if (!suspIns) {
            return error("development" === 'production', 'No Suspense instance');
        }
        var asyncFactorys = suspIns.asyncFactorys;
        if (error) {
            suspIns.$emit(REJECTED, error);
            return;
        }
        asyncFactorys.delete(asyncFactory);
        if (asyncFactorys.size === 0) {
            suspIns.$emit(RESOLVED);
        }
    };
    var add = function (asyncFactory) {
        var suspIns = currentSuspenseInstance || asyncFactory.suspenseInstance;
        if (!suspIns) {
            return error("development" === 'production', 'No Suspense instance');
        }
        var asyncFactorys = suspIns.asyncFactorys || (suspIns.asyncFactorys = new Set());
        if (suspIns.resolved) {
            suspIns.resolved = false;
            suspIns.setupLoading();
        }
        asyncFactorys.add(asyncFactory);
    };
    var has = function (asyncFactory) {
        var suspIns = currentSuspenseInstance || asyncFactory.suspenseInstance;
        if (!suspIns) {
            return error("development" === 'production', 'No Suspense instance');
        }
        return suspIns.asyncFactorys && suspIns.asyncFactorys.has(asyncFactory);
    };
    var Suspense = Vue.extend({
        name: COMPONENT_NAME,
        props: {
            delay: {
                type: Number,
                default: 0,
            },
        },
        data: function () {
            return {
                resloved: false,
                rejected: false,
                displayLoading: false,
            };
        },
        methods: {
            setupLoading: function () {
                var _this = this;
                if (this.delay > 0) {
                    this._time = setTimeout(function () {
                        _this.displayLoading = true;
                    }, this.delay);
                }
                else {
                    this.displayLoading = true;
                }
            },
            destoryLoading: function () {
                if (this._time) {
                    clearTimeout(this._time);
                    this._time = null;
                }
                this.displayLoading = false;
            },
        },
        created: function () {
            var _this = this;
            pushSuspenseInstance(this);
            this.promiser = new Promise(function (reslove, reject) {
                _this.$on(RESOLVED, function () {
                    _this.destoryLoading();
                    _this.resloved = true;
                    reslove();
                });
                _this.$on(REJECTED, function (err) {
                    _this.destoryLoading();
                    _this.rejected = true;
                    reject(err);
                });
            });
            // start loading
            this.setupLoading();
        },
        mounted: function () {
            if (!this.asyncFactorys) {
                this.$emit(RESOLVED);
            }
        },
        updated: function () {
            if (!this.resloved)
                return;
            popSuspenseInstance();
        },
        render: function (h) {
            var isVisible = this.$options.suspense.mode === 'visible';
            var emptyVNode = this._e();
            var fallback = this.displayLoading ? this.$slots.fallback || [emptyVNode] : [emptyVNode];
            // The `children` is the real content to be rendered
            var children = this.$slots.default || [emptyVNode];
            var rendered;
            if (this.rejected && this.$slots.error) {
                rendered = createWrapper(h, this.$slots.error);
            }
            else {
                rendered = isVisible
                    ? createWrapper(h, this.resolved ? children : children.concat(fallback))
                    : createWrapper(h, [
                        // We need to render the children, but we should not show the rendered content.
                        createHiddenWrapper(h, children, this.resolved),
                        fallback,
                    ]);
            }
            return rendered;
        },
    });
    function createWrapper(h, children) {
        return h('div', {
            class: { 'vue-suspense-wrapper': true },
        }, children);
    }
    function createHiddenWrapper(h, tree, display) {
        return h('div', {
            style: { display: display ? 'block' : 'none' },
            class: { 'vue-suspense-hidden-wrapper': true },
        }, tree);
    }

    function findSuspenseInstance(ins) {
        var current = ins.$parent;
        while (current) {
            if (current.$options.name === COMPONENT_NAME) {
                return current;
            }
            else {
                current = current.$parent;
            }
        }
        return;
    }

    function lazy(asyncFactory, props) {
        return Vue.extend({
            name: 'SuspenseLazy',
            props: props,
            created: function () {
                var _this = this;
                asyncFactory.suspenseInstance = currentSuspenseInstance || findSuspenseInstance(this);
                if (has(asyncFactory))
                    return;
                add(asyncFactory);
                if (asyncFactory.resolved) {
                    asyncFactory.$$waiter.then(function () {
                        del(asyncFactory);
                    });
                    return;
                }
                var promise = asyncFactory();
                asyncFactory.$$waiter = promise;
                promise
                    .then(function (C) {
                    // Compatible ES Module
                    if (C.__esModule && C.default) {
                        C = C.default;
                    }
                    asyncFactory.resolved = C;
                    // Trigger update
                    _this.$forceUpdate();
                })
                    .catch(function (err) {
                    error("development" === 'production', err);
                    del(asyncFactory, err);
                });
            },
            updated: function () {
                del(asyncFactory);
            },
            render: function (h) {
                var _this = this;
                // Fix context
                var slots = Object.keys(this.$slots)
                    .reduce(function (arr, key) { return arr.concat(_this.$slots[key] || []); }, [])
                    .map(function (vnode) {
                    vnode.context = _this._self;
                    return vnode;
                });
                return asyncFactory.resolved
                    ? h(asyncFactory.resolved, {
                        on: this.$listeners,
                        // If there is no props definition, fall back to `this.$attrs`
                        props: props ? this.$props : this.$attrs,
                        // Passthrough scopedSlots
                        scopedSlots: this.$scopedSlots,
                        attrs: this.$attrs,
                    }, slots)
                    : this._e();
            },
        });
    }

    Vue.mixin({
        created: function () {
            setCurrentInstance(this);
        },
    });
    //Vue.observable works in 2.6+
    function observable(data) {
        if (Vue.observable) {
            return Vue.observable(data);
        }
        return new Vue({
            data: function () {
                return data;
            },
        }).$data;
    }
    function createResource(fetchFactory, options) {
        var $res = observable({
            $$result: null,
            $$error: null,
            $$loading: false,
            $$loaded: false,
        });
        var currInstance = currentInstance;
        fetchFactory.suspenseInstance = currentSuspenseInstance
            ? currentSuspenseInstance
            : findSuspenseInstance(currInstance);
        var hasSuspenseInstance = !!fetchFactory.suspenseInstance;
        var resourceManager = {
            read: function (input) {
                // prevent
                if (options && options.prevent && $res.$$loading) {
                    return $res.$$promiser;
                }
                $res.$$loading = true;
                // Because we don't need caching, this is just a unique identifier,
                // and each call to .read() is a completely new request.
                var uniqueWrapFactory = function (i) {
                    return fetchFactory(i);
                };
                if (hasSuspenseInstance) {
                    // Establish a relationship between the fetchFactory and the current component instance
                    uniqueWrapFactory.suspenseInstance = fetchFactory.suspenseInstance;
                    add(uniqueWrapFactory);
                }
                // Start fetching asynchronous data
                var promise = ($res.$$promiser = uniqueWrapFactory(input));
                promise
                    .then(function (res) {
                    // Trigger update
                    $res.$$result = options && options.onSuccess ? options.onSuccess(res, input) : res;
                    if (!$res.$$loaded) {
                        $res.$$loaded = true;
                    }
                    if (hasSuspenseInstance) {
                        del(uniqueWrapFactory);
                    }
                })
                    .catch(function (err) {
                    error("development" === 'production', err.message);
                    $res.$$error = options && options.onError ? options.onError(err) : err;
                    if (hasSuspenseInstance) {
                        del(uniqueWrapFactory, err);
                    }
                })
                    .finally(function () {
                    $res.$$loading = false;
                });
                return promise;
            },
            get $result() {
                return $res.$$result;
            },
            set $result(val) {
                $res.$$result = val;
            },
            get $error() {
                return $res.$$error;
            },
            set $error(val) {
                $res.$$error = val;
            },
            get $loading() {
                return $res.$$loading;
            },
            get $loaded() {
                return $res.$$loaded;
            },
            fork: function () {
                return createResource(function (i) { return fetchFactory(i); }, options);
            },
        };
        return resourceManager;
    }

    function install(Vue, options) {
        if (options === void 0) { options = {}; }
        if (install.installed)
            return;
        install.installed = true;
        Vue.component('Suspense', Suspense);
        var opts = Object.assign({
            mode: 'visible',
        }, options);
        Vue.setSuspenseOptions = function (options) {
            Object.assign(opts, options);
        };
        Vue.mixin({
            created: function () {
                if (this.$options.name === COMPONENT_NAME) {
                    this.$options.suspense = opts;
                }
            },
        });
    }

    var index = {
        version: '0.1.1-alpha.3',
        install: install,
    };
    // Auto install if it is not done yet and `window` has `Vue`.
    // To allow users to avoid auto-installation in some cases,
    if (window && window.Vue) {
        window.Vue.use(install);
    }

    exports.Suspense = Suspense;
    exports.createReaourse = createResource;
    exports.default = index;
    exports.lazy = lazy;

    Object.defineProperty(exports, '__esModule', { value: true });

})));
